@@ -743,13 +743,13 @@ def _findprevtarget(repo, display, movebookmark=False, topic=True):
     else:
         header = _(b"multiple parents, choose one to update:")
         prevs = [p.rev() for p in parents]
-        choosedrev = utility.revselectionprompt(repo.ui, repo, prevs, header)
-        if choosedrev is None:
+        selectedrev = utility.revselectionprompt(repo.ui, repo, prevs, header)
+        if selectedrev is None:
             for p in parents:
                 display(p)
             repo.ui.warn(_(b'multiple parents, explicitly update to one\n'))
         else:
-            target = repo[choosedrev]
+            target = repo[selectedrev]
     return target, bookmark
 
 @eh.command(
@@ -845,9 +845,9 @@ def cmdnext(ui, repo, **opts):
         topic = _getcurrenttopic(repo)
         filtered = set()
         template = shorttemplate
-        if topic and not opts.get("no_topic", False):
-            filtered = set(ctx for ctx in children if ctx.topic() != topic)
-            children = [ctx for ctx in children if ctx not in filtered]
+        if topic and not opts['no_topic']:
+            filtered = set(ctx.rev() for ctx in children if ctx.topic() != topic)
+            children = [ctx for ctx in children if ctx.rev() not in filtered]
             template = utility.stacktemplate
             opts['stacktemplate'] = True
         display = compat.format_changeset_summary_fn(ui, repo, b'next',
@@ -856,10 +856,10 @@ def cmdnext(ui, repo, **opts):
         # check if we need to evolve while updating to the next child revision
         needevolve = False
         aspchildren = evolvecmd._aspiringchildren(repo, [repo[b'.'].rev()])
-        if topic:
-            filtered.update(repo[c] for c in aspchildren
-                            if repo[c].topic() != topic)
-            aspchildren = [ctx for ctx in aspchildren if ctx not in filtered]
+        if topic and not opts['no_topic']:
+            filtered.update(rev for rev in aspchildren
+                            if repo[rev].topic() != topic)
+            aspchildren = [rev for rev in aspchildren if rev not in filtered]
 
         # To catch and prevent the case when `next` would get confused by split,
         # lets filter those aspiring children which can be stablized on one of
@@ -869,7 +869,7 @@ def cmdnext(ui, repo, **opts):
             possdests = evolvecmd._possibledestination(repo, aspchild)
             if possdests & aspirants:
                 filtered.add(aspchild)
-        aspchildren = [ctx for ctx in aspchildren if ctx not in filtered]
+        aspchildren = [rev for rev in aspchildren if rev not in filtered]
         if aspchildren:
             needevolve = True
 
@@ -885,21 +885,21 @@ def cmdnext(ui, repo, **opts):
             else:
                 cmdutil.bailifchanged(repo, hint=_(b'do you want --merge?'))
 
-        if len(children) == 1:
+        if len(children) == 1 and (not opts['evolve'] or not aspchildren):
             c = children[0]
             return _updatetonext(ui, repo, c, display, opts)
         elif children:
             cheader = _(b"ambiguous next changeset, choose one to update:")
-            crevs = [c.rev() for c in children]
-            choosedrev = utility.revselectionprompt(ui, repo, crevs, cheader)
-            if choosedrev is None:
+            crevs = [c.rev() for c in children] + aspchildren
+            selectedrev = utility.revselectionprompt(ui, repo, crevs, cheader)
+            if selectedrev is None:
                 ui.warn(_(b"ambiguous next changeset:\n"))
-                for c in children:
-                    display(c)
+                for rev in crevs:
+                    display(repo[rev])
                 ui.warn(_(b"explicitly update to one of them\n"))
                 return 1
             else:
-                return _updatetonext(ui, repo, repo[choosedrev], display, opts)
+                return _updatetonext(ui, repo, repo[selectedrev], display, opts)
         else:
             if not opts['evolve'] or not aspchildren:
                 if filtered:
@@ -915,16 +915,16 @@ def cmdnext(ui, repo, **opts):
             elif len(aspchildren) > 1:
                 cheader = _(b"ambiguous next (unstable) changeset, choose one to"
                             b" evolve and update:")
-                choosedrev = utility.revselectionprompt(ui, repo,
-                                                        aspchildren, cheader)
-                if choosedrev is None:
+                selectedrev = utility.revselectionprompt(ui, repo,
+                                                         aspchildren, cheader)
+                if selectedrev is None:
                     ui.warn(_(b"ambiguous next (unstable) changeset:\n"))
                     for c in aspchildren:
                         display(repo[c])
                     ui.warn(_(b"(run 'hg evolve --rev REV' on one of them)\n"))
                     return 1
                 else:
-                    return _nextevolve(ui, repo, repo[choosedrev], opts)
+                    return _nextevolve(ui, repo, repo[selectedrev], opts)
             else:
                 return _nextevolve(ui, repo, aspchildren[0], opts)
     finally:
