@@ -6,7 +6,6 @@
 Compatibility module
 """
 
-import array
 import contextlib
 
 from mercurial import (
@@ -17,7 +16,6 @@ from mercurial import (
     logcmdutil,
     merge as mergemod,
     obsolete,
-    pycompat,
     registrar,
     repair,
     scmutil,
@@ -25,13 +23,6 @@ from mercurial import (
     ui as uimod,
 )
 
-if pycompat.ispy3:
-    arraytobytes = array.array.tobytes
-    arrayfrombytes = array.array.frombytes
-else:
-    arraytobytes = array.array.tostring
-    arrayfrombytes = array.array.fromstring
-
 # hg <= 5.2 (c21aca51b392)
 try:
     from mercurial import pathutil
