@@ -32,6 +32,8 @@ from mercurial import (
     util,
 )
 
+from mercurial.utils import stringutil
+
 from mercurial.i18n import _
 
 from . import (
@@ -164,7 +166,7 @@ def _solveorphan(ui, repo, orig, evolvestate, display, dryrun=False,
             progresscb()
         with state.saver(evolvestate, {b'current': orig.node()}):
             newid = _relocate(repo, orig, target, evolvestate, pctx,
-                              keepbranch, b'orphan')
+                              keepbranch, b'orphan', update=False)
             return (True, newid)
 
 def _solvephasedivergence(ui, repo, bumped, evolvestate, display,
@@ -316,7 +318,7 @@ def _pickresolutionparent(ui, repo, divergent, other, base, evolvestate):
 
     otherp1 = succsotherp1 = other.p1().rev()
     divp1 = succsdivp1 = divergent.p1().rev()
-    basep1 = base.p1().rev()
+    basep1 = succsbasep1 = base.p1().rev()
 
     # finding single successors of divp1 and otherp1
     try:
@@ -334,7 +336,7 @@ def _pickresolutionparent(ui, repo, divergent, other, base, evolvestate):
         ui.write_err(msg)
         return (False, b".")
     try:
-        utility._singlesuccessor(repo, base.p1())
+        succsbasep1 = utility._singlesuccessor(repo, base.p1())
     except utility.MultipleSuccessorsError:
         msg = (b"ambiguous orphan resolution parent for "
                b"%s (base)\n") % base
@@ -343,43 +345,41 @@ def _pickresolutionparent(ui, repo, divergent, other, base, evolvestate):
     # the changeset on which resolution changeset will be based on
     resolutionparent = succsdivp1
 
-    # divonly: non-obsolete csets which are topological ancestor of "divergent"
-    # but not "other"
-    divonly = repo.revs(b"only(%d, %d) - obsolete()" % (divergent.rev(),
-                                                        other.rev()))
-    # otheronly: non-obsolete csets which are topological ancestor of "other"
-    # but not "div"
-    otheronly = repo.revs(b"only(%d, %d) - obsolete()" % (other.rev(),
-                                                          divergent.rev()))
-    # make it exclusive set
-    divonly = set(divonly) - {divergent.rev()}
-    otheronly = set(otheronly) - {other.rev()}
-
     # testing how both the divergent changesets are arranged, there can be 4
     # possible cases here:
     #
-    # 1) both have the same parents
-    # 2) both have different parents but greatest common anscestor of them is
-    #    parent of one of them
-    # 3) both have different parents and gca is not parent of any of them
-    # 4) one of them is parent of other
+    # 1) both have the same parent (or parent's successor)
+    # 2) one of them is parent of other
+    # 3) one of them moved elsewhere
+    # 4) both of them moved elsewhere
     #
-    # we are handling 1) very good now.
-    # for 2) we will relocate one which is behind to the parent of ahead one and
-    # then solve the content-divergence the way we solve 1)
-    # for 3) and 4), we still have to decide
+    # we are handling 1) 2) 3) very well now.
+    # for 4), we still have to decide
     if otherp1 == divp1:
         # both are on the same parents
         pass
     elif divergent == other.p1():
         # both are in parent-child relation
         pass
-    elif basep1 in (divp1, otherp1):
-        # only one side moved, set parent of moved one as resolution parent
-        if basep1 == divp1:
+    elif succsbasep1 in (succsdivp1, succsotherp1):
+        # 1) either, only one side moved
+        #                -> set parent of moved one as resolution parent
+        # 2) or, both moved but one moved to succs of basep1
+        #                -> set parent of other one as resolution parent
+        # 3) or, both have same parent's successor
+        #                -> set parent's successor as resolution parent
+        if succsbasep1 == succsdivp1:
             resolutionparent = succsotherp1
         else:
             pass
+        # a special case of 3) where parent of moved one is obsolete with a
+        # successor. Also, look at section 'D-A3.1' in troubles-handling.rst
+        minimal_resolution = (
+            ui.configbool(b'experimental', b'evolution.divergence-resolution-minimal')
+            and succsdivp1 == succsotherp1 and basep1 in (divp1, otherp1)
+        )
+        if minimal_resolution:
+            resolutionparent = otherp1 if basep1 == divp1 else divp1
     else:
         # the nullrev has to be handled specially because -1 is overloaded to both
         # mean nullrev (this meaning is used for the result of changectx.rev(), as
@@ -390,22 +390,7 @@ def _pickresolutionparent(ui, repo, divergent, other, base, evolvestate):
             gca = [nodemod.nullrev]
         else:
             gca = repo.revs(b"ancestor(%d, %d)" % (succsotherp1, succsdivp1))
-
-        if succsotherp1 in gca and succsdivp1 in gca:
-            # both are not on the same parent but have same parents's succs.
-            if otheronly and divonly:
-                # case: we have visible csets on both side diverging from
-                # tca of "divergent" and "other". We still need to decide what
-                # to do in this case
-                pass
-            if otheronly:
-                resolutionparent = succsotherp1
-            elif divonly:
-                pass
-            else:
-                # no extra cset on either side
-                pass
-        elif succsotherp1 in gca and succsdivp1 not in gca:
+        if succsotherp1 in gca and succsdivp1 not in gca:
             pass
         elif succsdivp1 in gca and succsotherp1 not in gca:
             resolutionparent = succsotherp1
@@ -588,11 +573,11 @@ def _mergecontentdivergents(repo, progresscb, local, other, base,
     assert local != other
     if local not in repo[None].parents():
         repo.ui.note(_(b"updating to \"local\" side of the conflict: %s\n") %
-                     local.hex()[:12])
+                     local)
         compat.update(local)
     # merging the two content-divergent changesets
     repo.ui.note(_(b"merging \"other\" %s changeset '%s'\n") %
-                 (TROUBLES['CONTENTDIVERGENT'], other.hex()[:12]))
+                 (TROUBLES['CONTENTDIVERGENT'], other))
     if progresscb:
         progresscb()
     with state.saver(evolvestate):
@@ -896,29 +881,18 @@ def _orderrevs(repo, revs):
     ordering.extend(sorted(dependencies))
     return ordering
 
-def _relocate(repo, orig, dest, evolvestate, pctx=None, keepbranch=False,
-              category=None):
-    """rewrites the orig rev on dest rev
+def _rewrite_commit_message_hashes(repo, commitmsg):
+    """filter a commit description to update has to their successors
 
-    returns the node of new commit which is formed
+    The goal of this function is to avoid description referencing obsolete
+    hashes when a stack is rewritten or evolved.
+
+    Each hash in the description will be detected and, if matching an obsolete
+    changeset, it will be replaced by its successors.
+
+    Note: They might be case were such behavior might be is wrong, for example
+    if the commit message is explicitely referencing an older, obsolete changesets.
     """
-    if orig.rev() == dest.rev():
-        msg = _(b'tried to relocate a node on top of itself')
-        hint = _(b"This shouldn't happen. If you still need to move changesets, "
-                 b"please do so manually with nothing to rebase - working "
-                 b"directory parent is also destination")
-        raise error.ProgrammingError(msg, hint=hint)
-
-    if pctx is None:
-        if len(orig.parents()) == 2:
-            msg = _(b"tried to relocate a merge commit without specifying which "
-                    b"parent should be moved")
-            hint = _(b"Specify the parent by passing in pctx")
-            raise error.ProgrammingError(msg, hint)
-        pctx = orig.p1()
-
-    commitmsg = orig.description()
-
     cache = {}
     sha1s = re.findall(sha1re, commitmsg)
     unfi = repo.unfiltered()
@@ -940,17 +914,87 @@ def _relocate(repo, orig, dest, evolvestate, pctx=None, keepbranch=False,
         else:
             repo.ui.note(_(b'The stale commit message reference to %s could '
                            b'not be updated\n') % sha1)
+    return commitmsg
+
+def use_in_memory_merge(repo):
+    try:
+        from mercurial import mergestate as mergestatemod
+        mergestatemod.memmergestate
+    except (AttributeError, ImportError):
+        # no in-memory evolve if Mercurial lacks the required code
+        # hg <= 5.5 (19590b126764)
+        return False
+    config_value = repo.ui.config(b'experimental', b'evolution.in-memory')
+    if config_value == b'force':
+        return True
+    if repo.ui.hasconfig(b'hooks', b'precommit'):
+        return False
+    return stringutil.parsebool(config_value)
+
+def _relocate(repo, orig, dest, evolvestate, pctx=None, keepbranch=False,
+              category=None, update=True):
+    """rewrites the orig rev on dest rev
+
+    Also updates bookmarks and creates obsmarkers.
+
+    returns the node of new commit which is formed
+    """
+    if orig.rev() == dest.rev():
+        msg = _(b'tried to relocate a node on top of itself')
+        hint = _(b"This shouldn't happen. If you still need to move changesets, "
+                 b"please do so manually with nothing to rebase - working "
+                 b"directory parent is also destination")
+        raise error.ProgrammingError(msg, hint=hint)
+
+    if pctx is None:
+        if len(orig.parents()) == 2:
+            msg = _(b"tried to relocate a merge commit without specifying which "
+                    b"parent should be moved")
+            hint = _(b"Specify the parent by passing in pctx")
+            raise error.ProgrammingError(msg, hint)
+        pctx = orig.p1()
+
+    commitmsg = _rewrite_commit_message_hashes(repo, orig.description())
 
     tr = repo.currenttransaction()
     assert tr is not None
     if repo._activebookmark:
         repo.ui.status(_(b"(leaving bookmark %s)\n") % repo._activebookmark)
     bookmarksmod.deactivate(repo)
-    nodenew = _relocatecommit(repo, orig, dest, pctx, keepbranch, commitmsg)
+    nodenew = _relocatecommit(repo, orig, dest, pctx, keepbranch, commitmsg,
+                              update)
     _finalizerelocate(repo, orig, dest, nodenew, tr, category, evolvestate)
     return nodenew
 
-def _relocatecommit(repo, orig, dest, pctx, keepbranch, commitmsg):
+def _relocatecommit(repo, orig, dest, pctx, keepbranch, commitmsg, update):
+    extra = dict(orig.extra())
+    if b'branch' in extra:
+        del extra[b'branch']
+    extra[b'rebase_source'] = orig.hex()
+    targetphase = max(orig.phase(), phases.draft)
+    configoverrides = {
+        (b'phases', b'new-commit'): targetphase
+    }
+    with repo.ui.configoverride(configoverrides, source=b'evolve'):
+        if not update and use_in_memory_merge(repo):
+            try:
+                return _relocatecommitinmem(
+                    repo, orig, dest, pctx, keepbranch, commitmsg, extra
+                )
+            except error.InMemoryMergeConflictsError:
+                repo.ui.status(
+                    b'hit merge conflicts; retrying merge in working copy\n')
+        return _relocatecommitondisk(repo, orig, dest, pctx, keepbranch,
+                                     commitmsg, extra)
+
+def _relocatecommitondisk(repo, orig, dest, pctx, keepbranch, commitmsg, extra):
+    """rewrites the orig rev on dest rev on disk
+
+    Creates the new commit by using the old-fashioned way of creating
+    commits by writing files to the working copy.
+
+    returns the node of new commit which is formed
+    """
     if repo[b'.'].rev() != dest.rev():
         compat._update(repo, dest, branchmerge=False, force=True)
     if keepbranch:
@@ -971,19 +1015,50 @@ def _relocatecommit(repo, orig, dest, pctx, keepbranch, commitmsg):
         raise error.InterventionRequired(_(b"unresolved merge conflicts"),
                                          hint=hint)
 
-    extra = dict(orig.extra())
-    if b'branch' in extra:
-        del extra[b'branch']
-    extra[b'rebase_source'] = orig.hex()
+    # Commit might fail if unresolved files exist
+    return repo.commit(text=commitmsg, user=orig.user(),
+                       date=orig.date(), extra=extra)
 
-    targetphase = max(orig.phase(), phases.draft)
-    configoverride = repo.ui.configoverride({
-        (b'phases', b'new-commit'): targetphase
-    }, source=b'evolve')
-    with configoverride:
-        # Commit might fail if unresolved files exist
-        return repo.commit(text=commitmsg, user=orig.user(),
-                           date=orig.date(), extra=extra)
+def _relocatecommitinmem(repo, orig, dest, pctx, keepbranch, commitmsg, extra):
+    """rewrites the orig rev on dest rev in memory
+
+    Creates the new commit by using the modern way of creating
+    commits without writing files to the working copy.
+
+    returns the node of new commit which is formed
+    """
+    wctx = context.overlayworkingctx(repo)
+    wctx.setbase(dest)
+
+    stats = merge.graft(
+        repo,
+        orig,
+        pctx,
+        [b'destination', b'evolving'],
+        keepparent=True,
+        wctx=wctx
+    )
+
+    if stats.unresolvedcount: # some conflict
+        raise error.InMemoryMergeConflictsError()
+
+    branch = None
+    if keepbranch:
+        branch = orig.branch()
+    # FIXME: We call _compact() because it's required to correctly detect
+    # changed files. This was added to fix a regression shortly before the 5.5
+    # release. A proper fix will be done in the default branch.
+    wctx._compact()
+    memctx = wctx.tomemctx(
+        commitmsg,
+        date=orig.date(),
+        extra=extra,
+        user=orig.user(),
+        branch=branch,
+    )
+    if memctx.isempty() and not repo.ui.configbool(b'ui', b'allowemptycommit'):
+        return None
+    return repo.commitctx(memctx)
 
 def _finalizerelocate(repo, orig, dest, nodenew, tr, category, evolvestate):
     if nodenew is not None:
@@ -1397,8 +1472,20 @@ def _checkevolveopts(repo, opts):
 
     return opts
 
-def _cleanup(ui, repo, startnode, shouldupdate):
-    if not shouldupdate:
+def _cleanup(ui, repo, startnode, shouldupdate, headnode):
+    """Update to the right destination after evolving, if necessary
+
+    headnode is the last node created by the evolve operation (which
+    we may need to update to when using in-memory merge)
+    """
+    # If --update was passed, we should update to some head of the evolved set,
+    # but we only need to do that in the in-memory case. If --update was not
+    # passed, we should still update the working copy to its successor if there
+    # is one.
+    if shouldupdate:
+        if use_in_memory_merge(repo) and headnode:
+            compat.update(repo[headnode])
+    else:
         # Move back to startnode, or to its successor if the start node is
         # obsolete (perhaps made obsolete by the current `hg evolve`)
         unfi = repo.unfiltered()
@@ -1593,13 +1680,14 @@ def _performevolve(ui, repo, **opts):
 
     ui.setconfig(b'ui', b'forcemerge', opts.get('tool', r''), b'evolve')
 
+    headnode = None
     evolvestate = state.cmdstate(repo)
     # Continuation handling
     if contopt:
         if not evolvestate:
             raise error.Abort(_(b'no interrupted evolve to continue'))
         evolvestate.load()
-        continueevolve(ui, repo, evolvestate)
+        headnode = continueevolve(ui, repo, evolvestate)
         if evolvestate[b'command'] != b'evolve':
             evolvestate.delete()
             return
@@ -1612,8 +1700,11 @@ def _performevolve(ui, repo, **opts):
             raise error.Abort(_(b'no interrupted evolve to stop'))
         evolvestate.load()
         stopevolve(ui, repo, evolvestate)
+        if evolvestate[b'command'] != b'evolve':
+            evolvestate.delete()
+            return
+        startnode = repo.unfiltered()[evolvestate[b'startnode']]
         evolvestate.delete()
-        return
     elif abortopt:
         if not evolvestate:
             raise error.Abort(_(b'no interrupted evolve to abort'))
@@ -1624,7 +1715,7 @@ def _performevolve(ui, repo, **opts):
             compat.clean_update(pctx)
             ui.status(_(b'evolve aborted\n'))
             ui.status(_(b'working directory is now at %s\n')
-                      % pctx.hex()[:12])
+                      % pctx)
             evolvestate.delete()
             return 0
         return abortevolve(ui, repo, evolvestate)
@@ -1672,21 +1763,26 @@ def _performevolve(ui, repo, **opts):
         tr = repo.transaction(b"evolve")
         with util.acceptintervention(tr):
             for rev in revs:
-                _solveonerev(ui, repo, rev, evolvestate, activetopic, dryrunopt,
-                             confirmopt, progresscb, targetcat)
+                (solved, newnode) = _solveonerev(ui, repo, rev, evolvestate,
+                                                 activetopic, dryrunopt,
+                                                 confirmopt, progresscb,
+                                                 targetcat)
+                if solved:
+                    headnode = newnode
                 seen += 1
 
         if showprogress:
             compat.progress(ui, _(b'evolve'), None)
 
-    _cleanup(ui, repo, startnode, shouldupdate)
+    _cleanup(ui, repo, startnode, shouldupdate, headnode)
 
 def _solveonerev(ui, repo, rev, evolvestate, activetopic, dryrunopt, confirmopt,
                  progresscb, targetcat):
     """solves one trouble, including orphan merges
 
     Like _solveone(), this solves one trouble. Unlike _solveone(), it
-    stabilizes for both parents of orphan merges.
+    stabilizes for both parents of orphan merges. Returns the same value as
+    _solveone().
     """
     curctx = repo[rev]
     revtopic = getattr(curctx, 'topic', lambda: b'')()
@@ -1719,6 +1815,7 @@ def _solveonerev(ui, repo, rev, evolvestate, activetopic, dryrunopt, confirmopt,
             evolvestate[b'skippedrevs'].append(curctx.node())
 
         evolvestate[b'orphanmerge'] = False
+    return ret
 
 def solveobswdp(ui, repo, opts):
     """this function updates to the successor of obsolete wdir parent"""
@@ -1787,7 +1884,6 @@ def stopevolve(ui, repo, evolvestate):
         pctx = repo[b'.']
         compat.clean_update(pctx)
     ui.status(_(b'stopped the interrupted evolve\n'))
-    ui.status(_(b'working directory is now at %s\n') % pctx)
 
 def abortevolve(ui, repo, evolvestate):
     """ logic for handling of `hg evolve --abort`"""
@@ -1862,7 +1958,7 @@ def abortevolve(ui, repo, evolvestate):
         evolvestate.delete()
         ui.status(_(b'evolve aborted\n'))
         ui.status(_(b'working directory is now at %s\n')
-                  % nodemod.hex(startnode)[:12])
+                  % nodemod.short(startnode))
     else:
         raise error.Abort(_(b"unable to abort interrupted evolve, use 'hg "
                             b"evolve --stop' to stop evolve"))
@@ -1877,7 +1973,7 @@ def hgabortevolve(ui, repo):
             compat.clean_update(pctx)
             ui.status(_(b'evolve aborted\n'))
             ui.status(_(b'working directory is now at %s\n')
-                      % pctx.hex()[:12])
+                      % pctx)
             evolvestate.delete()
             return 0
         return abortevolve(ui, repo, evolvestate)
@@ -1911,6 +2007,7 @@ def continueevolve(ui, repo, evolvestate):
         compat.progress(ui, _(b'evolve'), seen, unit=_(b'changesets'),
                         total=count)
 
+    headnode = None
     category = evolvestate[b'category']
     confirm = evolvestate[b'confirm']
     unfi = repo.unfiltered()
@@ -1937,9 +2034,11 @@ def continueevolve(ui, repo, evolvestate):
                 if newnode[0]:
                     evolvestate[b'replacements'][curctx.node()] = newnode[1]
                     evolvestate[b'lastsolved'] = newnode[1]
+                    headnode = newnode[1]
                 else:
                     evolvestate[b'skippedrevs'].append(curctx.node())
             seen += 1
+    return headnode
 
 def _continuecontentdivergent(ui, repo, evolvestate, progresscb):
     """function to continue the interrupted content-divergence resolution."""
