@@ -22,6 +22,7 @@ import sqlite3
 import struct
 import weakref
 
+from mercurial.i18n import _
 from mercurial import (
     encoding,
     error,
@@ -34,7 +35,6 @@ from mercurial import (
     store,
     util,
 )
-from mercurial.i18n import _
 
 from mercurial.utils.stringutil import forcebytestr
 
@@ -472,7 +472,7 @@ class _obshashcache(obscache.dualsourcecache):
         def progress(pos, rev=None):
             revstr = b'' if rev is None else (b'rev %d' % rev)
             compat.progress(repo.ui, b'updating obshashrange cache',
-                            pos, revstr, unit=b'revision', total=total)
+                            pos, revstr, unit=_(b'changesets'), total=total)
         # warm the cache for the new revs
         progress(0)
         for idx, r in enumerate(revs):
