@@ -61,7 +61,7 @@ def addtouchnoise(ui):
      (b'f', b'filternonlocal', False, _(b'filter out non local commits')),
      (b'o', b'origin', True, _(b'show origin of changesets instead of fate')),
      ] + commands.formatteropts,
-    _(b'hg olog [OPTION]... [[-r] REV]...'),
+    _(b'hg obslog [OPTION]... [[-r] REV]...'),
     **compat.helpcategorykwargs('CATEGORY_CHANGE_NAVIGATION'))
 def cmdobshistory(ui, repo, *revs, **opts):
     """show the obsolescence history of the specified revisions
@@ -219,9 +219,14 @@ def _fatemarkers(repo, ctx, filternonlocal):
         r = obsutil.successorsandmarkers(repo, ctx)
         if r is None:
             r = []
-        for succset in sorted(r):
-            if succset[b'markers']:
-                yield (succset[b'successors'], succset[b'markers'])
+        # replacing dicts and sets, they can't be compared
+        data = [
+            (succset[b'successors'], tuple(sorted(succset[b'markers'])))
+            for succset in r
+        ]
+        for (nodes, markers) in sorted(data):
+            if markers:
+                yield (nodes, markers)
     else:
         markers = successors.get(ctx.node(), ())
         for marker in sorted(markers):
