@@ -250,6 +250,7 @@ import array
 import collections
 import struct
 
+from mercurial.i18n import _
 from mercurial import (
     commands,
     error,
@@ -262,8 +263,6 @@ from mercurial import (
 
 from mercurial.utils.stringutil import forcebytestr
 
-from mercurial.i18n import _
-
 from . import (
     compat,
     depthcache,
@@ -337,11 +336,12 @@ def debugstablesortcache(ui, repo, **opts):
     unfi = repo.unfiltered()
     revs = unfi.revs('all()')
     nbrevs = len(revs)
-    ui.write('number of revisions: %12d\n' % nbrevs)
+    ui.write(b'number of revisions: %12d\n' % nbrevs)
     merge = unfi.revs('merge()')
     nbmerge = len(merge)
     cache = unfi.stablesort
-    ui.write('number of merge:     %12d\n' % nbmerge)
+    cache.save(repo)
+    ui.write(b'number of merge:     %12d\n' % nbmerge)
     alljumps = []
     alljumpssize = []
     for r in merge:
@@ -352,20 +352,20 @@ def debugstablesortcache(ui, repo, **opts):
         alljumps.append(jumps)
         alljumpssize.append(len(jumps))
     nbjumps = sum(alljumpssize)
-    ui.write('number of jumps:     %12d\n' % nbjumps)
+    ui.write(b'number of jumps:     %12d\n' % nbjumps)
     if not nbjumps:
         return 0
     avgjumps = nbjumps / float(len(alljumpssize))
-    ui.write('average jumps:                 %6.3f\n' % avgjumps)
+    ui.write(b'average jumps:                 %6.3f\n' % avgjumps)
     alljumpssize.sort()
     medianjumps = alljumpssize[len(alljumpssize) // 2]
-    ui.write('median jumps:        %12d\n' % medianjumps)
+    ui.write(b'median jumps:        %12d\n' % medianjumps)
     tensjumps = alljumpssize[len(alljumpssize) * 9 // 10]
-    ui.write('90%% jumps:           %12d\n' % tensjumps)
+    ui.write(b'90%% jumps:           %12d\n' % tensjumps)
     centsjumps = alljumpssize[len(alljumpssize) * 99 // 100]
-    ui.write('99%% jumps:           %12d\n' % centsjumps)
-    ui.write('max jumps:           %12d\n' % max(alljumpssize))
-    ui.write('jump cache size:     %12d bytes\n' % (nbjumps * 12))
+    ui.write(b'99%% jumps:           %12d\n' % centsjumps)
+    ui.write(b'max jumps:           %12d\n' % max(alljumpssize))
+    ui.write(b'jump cache size:     %12d bytes\n' % (nbjumps * 12))
 
 def stablesort_branchpoint(repo, revs, mergecallback=None):
     """return '::revs' topologically sorted in "stable" order
@@ -864,7 +864,7 @@ class ondiskstablesortcache(stablesortcache, genericcaches.changelogsourcebase):
         def progress(pos, rev=None):
             revstr = b'' if rev is None else (b'rev %d' % rev)
             compat.progress(repo.ui, b'updating stablesort cache',
-                            pos, revstr, unit=b'revision', total=total)
+                            pos, revstr, unit=_(b'changesets'), total=total)
 
         progress(0)
         for idx, rev in enumerate(data):
@@ -895,20 +895,31 @@ class ondiskstablesortcache(stablesortcache, genericcaches.changelogsourcebase):
         assert repo.filtername is None
 
         data = repo.cachevfs.tryread(self._filepath)
+        self._cachekey = self.emptykey
         self._index = array.array(r'l')
         self._data = array.array(r'l')
-        if not data:
-            self._cachekey = self.emptykey
-        else:
+        if data:
             headerdata = data[:self._cachekeysize]
-            self._cachekey = self._deserializecachekey(headerdata)
+            cachekey = self._deserializecachekey(headerdata)
             offset = self._cachekeysize
             indexsizedata = data[offset:offset + S_INDEXSIZE.size]
             indexsize = S_INDEXSIZE.unpack(indexsizedata)[0]
             offset += S_INDEXSIZE.size
-            compat.arrayfrombytes(self._index, data[offset:offset + indexsize])
-            offset += indexsize
-            compat.arrayfrombytes(self._data, data[offset:])
+            if indexsize % self._datastruct.size == 0 and len(data) - offset >= indexsize:
+                index = list(self._deserializedata(data[offset:offset + indexsize]))
+                offset += indexsize
+                expected = index[-1] * self._datastruct.size * 3
+                data = data[offset:]
+            else:
+                # index cannot be read, so we need to abort somehow
+                expected = None
+            if len(data) == expected:
+                self._index.extend(index)
+                self._data.extend(self._deserializedata(data))
+                self._cachekey = cachekey
+            else:
+                repo.ui.debug(b'stablesortcache file seems to be corrupted, '
+                              b'it will be rebuilt from scratch\n')
         self._ondiskkey = self._cachekey
         pass
 
@@ -924,8 +935,8 @@ class ondiskstablesortcache(stablesortcache, genericcaches.changelogsourcebase):
 
             # data to write
             headerdata = self._serializecachekey()
-            indexdata = compat.arraytobytes(self._index)
-            data = compat.arraytobytes(self._data)
+            indexdata = self._serializedata(self._index)
+            data = self._serializedata(self._data)
             indexsize = S_INDEXSIZE.pack(len(indexdata))
 
             # writing
