@@ -232,9 +232,9 @@ colortable = {b'topic.active': b'green',
               b'topic.active': b'green',
               }
 
-__version__ = b'0.22.0'
+__version__ = b'0.22.1'
 
-testedwith = b'4.6.2 4.7 4.8 4.9 5.0 5.1 5.2 5.3 5.4 5.5 5.6 5.7'
+testedwith = b'4.6.2 4.7 4.8 4.9 5.0 5.1 5.2 5.3 5.4 5.5 5.6 5.7 5.8'
 minimumhgversion = b'4.6'
 buglink = b'https://bz.mercurial-scm.org/'
 
